/*

Parts adapted from Adafruit TFT display example code.
    
Copyright (c) 2017 Insight Machines Laboratory

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*/

#include <Adafruit_GFX.h>
#include <Adafruit_ST7735.h> 
#include <Fonts/FreeSans18pt7b.h>
#include <Fonts/FreeSans9pt7b.h>
#include <SD.h>
#include <Wire.h>
#include "RTClib.h"
#include "SparkFunHTU21D.h"
#include "Adafruit_VEML6070.h"

#define BLACK       0x0000
#define BLUE        0x001F
#define RED         0xF800
#define GREEN       0x07E0
#define CYAN        0x07FF
#define MAGENTA     0xF81F
#define YELLOW      0xFFE0  
#define WHITE       0xFFFF

struct Data {
   int temperature; // [deg C]
   int humidity;    // [%]
   int uvLevel;     // []
   int pm2_5;       // [ug/m3]
   int pm10;        // [ug/m3]
};

// Pins
const int PIN_DISPLAY_CS = 10;
const int PIN_DISPLAY_RST = 13;
const int PIN_DISPLAY_DC = 6;
const int PIN_SD_CS = 5;
const int PIN_BUZZER = A4;
const int PIN_BUTTON = 11;


// Times/durations/intervals
const int UPDATE_INTERVAL = 1000; // [ms]
const int DEBOUNCE_DELAY = 50;   // [ms]

// Screen
const int BACKGROUND = BLACK;
const int WIDTH = 128;  // [px]
const int HEIGHT = 128; // [px]
const int NUM_SCREENS = 5;

// PM Sensor
const int PM_MESSAGE_LENGTH = 30;
unsigned char pmMessageBuffer[PM_MESSAGE_LENGTH];
const char PM_MESSAGE_START_TOKEN[2] = {(char) 0x42, (char) 0x4d};
const int PM_HEADER_SUM = 0x42+0x4d;
const int PM_1_0_INDEX = 8;  // Currently unused
const int PM_2_5_INDEX = 10;
const int PM_10_INDEX = 12;

enum Command { NONE, SWITCH_SCREEN };

Adafruit_ST7735 tft = Adafruit_ST7735(PIN_DISPLAY_CS, PIN_DISPLAY_DC, PIN_DISPLAY_RST);
Adafruit_VEML6070 veml6070 = Adafruit_VEML6070();
HTU21D htu21d;

void (*screenRenderers[])(bool) = { &renderTemperature, &renderHumidity, &renderUVLevel, &renderPM2_5, &renderPM10 };
unsigned long nextDataUpdateTime = 0;  // [ms]
unsigned long lastDebounceTime = 0;  // [ms]
unsigned long buttonPressedTime = 0;  // [ms]

int lastButtonState = HIGH;
int buttonState;
int screenIndex;
Data currentData;
Data previousData;
char valueChars[6];


void setup() {
    Serial.begin(9600);    
    setupSdCard();
    setupDisplay();
    renderSplashScreen();    
    setupScreenControl();
    setupSensors();    
    playPostSound();
    clearScreen();
}

void setupSdCard() {
  Serial.println("-- Initializing SD card --");
  if (!SD.begin(PIN_SD_CS)) {
    Serial.println("Failed!");
    return;
  }
  Serial.println("OK!");
}

void setupDisplay() {
    Serial.println("-- Setting up display --");    
    tft.initR(INITR_144GREENTAB); 
    tft.fillScreen( BLACK );
    tft.setRotation(0);
    tft.setTextWrap(false);
    tft.setTextSize(1);
    tft.setTextColor(WHITE);    
}

void renderSplashScreen() {
    Serial.println("-- Rendering splash screen --");

    const char*    filename = "splash.bmp";
    const uint8_t  bmpWidth = 128; 
    const uint8_t  bmpHeight = 128;
    const uint32_t rowSize = 3*bmpWidth;

    File      bmpFile;
    uint8_t   bmpDepth;           // Bit depth (currently must be 24)
    uint32_t  bmpImageOffset;     // Start of image data in file
    uint8_t   sdbuffer[rowSize];  // pixel buffer (R+G+B per pixel)
    uint32_t  buffidx = rowSize;  // Current position in sdbuffer
    boolean   goodBmp = false;    // Set to true on valid header parse
    uint32_t  row, col;
    uint8_t   r, g, b;
    uint32_t  pos = 0;

    // Open requested file on SD card
    if ((bmpFile = SD.open(filename)) == NULL) {
        Serial.println("File not found");
        return;
    }

    // Parse BMP header
    if (read16(bmpFile) == 0x4D42) { // BMP signature
        (void) read32(bmpFile); // Read & ignore file size
        (void) read32(bmpFile); // Read & ignore creator bytes
        bmpImageOffset = read32(bmpFile); // Start of image data
        (void) read32(bmpFile); // Read & ignore DIB header
        (void) read32(bmpFile); // Read & ignore image width
        (void) read32(bmpFile); // Read & ignore image height
        if ( read16(bmpFile) == 1 ) { // # planes -- must be '1'
            bmpDepth = read16(bmpFile); // bits per pixel
            if ( (bmpDepth == 24) && (read32(bmpFile) == 0) ) { // 0 = uncompressed
                // Set TFT address window to image bounds
                tft.setAddrWindow(0, 0, bmpWidth-1, bmpHeight-1);            
                for (row=0; row<bmpHeight; row++) {
                    // Flip the position as it is stored bottom-to-top order:
                    pos = bmpImageOffset + (bmpHeight - 1 - row) * rowSize;
                    // Read next row:
                    bmpFile.seek(pos);
                    bmpFile.read(sdbuffer, rowSize);
                    buffidx = 0; // Set index to beginning
                    for (col=0; col<bmpWidth; col++) { // For each pixel...
                        // Convert pixel from BMP to TFT format, push to display
                        b = sdbuffer[buffidx++];
                        g = sdbuffer[buffidx++];
                        r = sdbuffer[buffidx++];
                        tft.pushColor(tft.Color565(r,g,b));
                    }
                }            
            } else {
                Serial.println("Bad color depth. Should be 24 bits, uncompressed.");
            }
        } else {
            Serial.println("Bad number of planes. Should be 1.");
        }
    }

    bmpFile.close();  
}

void setupScreenControl() {
    Serial.println("-- Setting up screen control --");
    pinMode(PIN_BUTTON, INPUT_PULLUP);
    screenIndex = -1;
}

void setupSensors() {
    Serial.println("-- Setting up sensors --");
    
    // Start I2C:
    Wire.begin();
    
    // Start the Temperature/Humidity sensor:
    htu21d.begin();
    
    // Start the UV sensor
    veml6070.begin(VEML6070_1_T);
    
    // Initialize UART connection with the PM sensor:
    Serial1.begin(9600);
    Serial1.setTimeout(1500); 
}

void playPostSound() {
    Serial.println("-- Setup complete --");
    pinMode(PIN_BUZZER, OUTPUT);
    tone(PIN_BUZZER, 800);  // Send 800Hz ping
    delay(500);
    noTone(PIN_BUZZER);
}

void clearScreen() {
    tft.fillScreen( BLACK );
}

void loop() {   
    unsigned long now = millis();
    Command command = getUserCommand(now);
    switch (command) {
        case SWITCH_SCREEN:
            showNextScreen();
            break;
        default:
            if ( screenIndex == -1 ) {
                showNextScreen();
            } else if ( now > nextDataUpdateTime ) {
                nextDataUpdateTime = now + UPDATE_INTERVAL;
                updateData();
                refreshCurrentScreen();
            }
    }       
}

Command getUserCommand( unsigned long now ) {
    int reading = digitalRead(PIN_BUTTON);

    if (reading != lastButtonState) {
        lastDebounceTime = millis();
    }
    lastButtonState = reading;

    if ((now - lastDebounceTime) > DEBOUNCE_DELAY) {
        if (reading != buttonState) {
            buttonState = reading;
            if (buttonState == HIGH) {
                return SWITCH_SCREEN;
            } else {
                buttonPressedTime = now;
            }
        }
    }
    
    return NONE;
}

void showNextScreen() {
    screenIndex = (screenIndex+1)%NUM_SCREENS;
    screenRenderers[screenIndex](true);
}

void refreshCurrentScreen() {
    screenRenderers[screenIndex](false);
}

void updateData() {
    updatePMBuffer();
    previousData = currentData;
    currentData.temperature = readTemperature();
    currentData.humidity = readHumidity();
    currentData.pm2_5 = readPM2_5();
    currentData.pm10 = readPM10();
    currentData.uvLevel = readUV();
}

int readTemperature() {
    return round( htu21d.readTemperature() );
}

int readHumidity() {
    return round( htu21d.readHumidity() );
}

void updatePMBuffer() {
    char * startToken = (char *) PM_MESSAGE_START_TOKEN;
    while ( Serial1.available() > PM_MESSAGE_LENGTH ) {
        if ( Serial1.find( startToken ) ) {
            Serial1.readBytes( pmMessageBuffer, PM_MESSAGE_LENGTH );
        }
    }
}

int readPM2_5() {
    return readPMValueFromBuffer( PM_2_5_INDEX );
}

int readPM10() {
    return readPMValueFromBuffer( PM_10_INDEX );
}

int readUV() {
    return veml6070.readUV();
}

bool checkPMMessage() {  
    int receiveSum = PM_HEADER_SUM;
    for(int i=0; i<(PM_MESSAGE_LENGTH-2); i++){
        receiveSum += pmMessageBuffer[i];
    }     
    return ( receiveSum == ((pmMessageBuffer[PM_MESSAGE_LENGTH-2]<<8)+pmMessageBuffer[PM_MESSAGE_LENGTH-1]) );
}

int readPMValueFromBuffer( int index ) {
    return ((pmMessageBuffer[index]<<8) + pmMessageBuffer[index+1]);
}

void renderTemperature( bool clearAll ) {
    renderScreen( clearAll, "Temperature", previousData.temperature, currentData.temperature, renderTemperatureValue );
}

void renderHumidity( bool clearAll ) {
    renderScreen( clearAll, "Humidity", previousData.humidity, currentData.humidity, renderHumidityValue );
}

void renderUVLevel( bool clearAll ) {
    renderScreen( clearAll, "UV Level", previousData.uvLevel, currentData.uvLevel, renderUVValue );
}

void renderPM2_5( bool clearAll ) {
    renderScreen( clearAll, "PM 2.5", previousData.pm2_5, currentData.pm2_5, renderPMValue );
}

void renderPM10( bool clearAll ) {
    renderScreen( clearAll, "PM 10", previousData.pm10, currentData.pm10, renderPMValue );
}

void renderTemperatureValue( int value, int color ) {
    int16_t x1, y1;
    uint16_t w, h;
    sprintf( valueChars, "%i", value );
    
    tft.setFont(&FreeSans18pt7b);
    tft.getTextBounds( valueChars, 0, 0, &x1, &y1, &w, &h );  // 23x8 px
    w += 3; // pixels space
    w += 8; // pixels for degree sign
    w += 2; // pixels space
    w += 26; // pixels for C
    
    x1 = (WIDTH-w)/2;
    y1 = (HEIGHT-h)/2;

    tft.setTextColor( color, BACKGROUND ); tft.setCursor( x1, y1+h ); tft.print( valueChars );    
    tft.fillCircle( x1+w-29, y1, 4, color );
    tft.fillCircle( x1+w-29, y1, 2, BACKGROUND );    
    tft.setCursor( x1+w-25, y1+h ); tft.print( "C" );
}

void renderHumidityValue( int value, int color ) {
    int16_t x1, y1;
    uint16_t w, h;
    sprintf( valueChars, "%i", value );
    
    tft.setFont(&FreeSans18pt7b);
    tft.getTextBounds( valueChars, 0, 0, &x1, &y1, &w, &h ); 
    w += 3; // pixels space
    w += 26; // pixels for %
    
    x1 = (WIDTH-w)/2;
    y1 = (HEIGHT-h)/2;
    tft.setTextColor( color, BACKGROUND ); tft.setCursor( x1, y1+h ); tft.print( valueChars );
    tft.setCursor( x1+w-25, y1+h ); tft.print( "%" );
}

void renderUVValue( int value, int color ) {
    int16_t x1, y1;
    uint16_t w, h;
    sprintf( valueChars, "%i", value );
        
    tft.setFont(&FreeSans18pt7b);
    tft.getTextBounds( valueChars, 0, 0, &x1, &y1, &w, &h ); 
    
    x1 = (WIDTH-w)/2;
    y1 = (HEIGHT-h)/2;
    tft.setTextColor( color, BACKGROUND ); tft.setCursor( x1, y1+h ); tft.print( valueChars );
}

void renderPMValue( int value, int color ) {
    static const int16_t unitsWidth = 23;
    static const int16_t unitsHeight = 37;
    
    int16_t x1, y1, x2, y2;
    uint16_t w, h;    
    sprintf( valueChars, "%i", value );
     
    tft.setFont(&FreeSans18pt7b);
    tft.getTextBounds( valueChars, 0, 0, &x1, &y1, &w, &h );
    w += 10; // pixels space
    w += unitsWidth; // pixels for ug/m3
    
    x1 = (WIDTH-w)/2;
    y1 = (HEIGHT-h)/2;
    tft.setTextColor( color, BACKGROUND ); tft.setCursor( x1, y1+h ); tft.print( valueChars );
    
    x2 = x1+w-unitsWidth;
    y2 = y1+h-20;
    tft.setFont(&FreeSans9pt7b);    
    tft.setCursor( x2, y2 ); tft.print( "ug" );
    tft.setCursor( x2+1, y2+25 ); tft.print( "m" );
    tft.setFont();
    tft.setCursor( x2+17, y2+12 ); tft.print( "3" );
    
    // Fraction line
    tft.drawFastHLine( x2, y2+8, 20, color );
    
    // Micro sign longer leg:
    tft.drawFastVLine( x2+1, y2-1, 5, color );
    tft.drawFastVLine( x2+2, y2-1, 5, color );
}

void renderHeader( const char* headerChars ) {
    int16_t x1, y1;
    uint16_t w, h;
    clearDisplayArea();
    tft.setFont();
    tft.getTextBounds( (char*) headerChars, 0, 0, &x1, &y1, &w, &h );
    tft.setCursor((WIDTH-w)/2,30); tft.setTextColor( WHITE, BACKGROUND ); tft.print( headerChars );
}

void renderScreen( bool clearAll, const char* headerChars, int previousValue, int currentValue, void (*renderValue)(int, int) ) {
    if ( clearAll ) {
        renderHeader( headerChars );
    } else {
        if ( currentValue == previousValue ) return;    
        renderValue( previousValue, BACKGROUND );
    }
    renderValue( currentValue, CYAN );
}

void clearDisplayArea() {
    tft.fillRect( 0, 30, 128, 55, BACKGROUND );
}

// *****************************
//       UTILITY FUNCTIONS
// *****************************

uint16_t read16(File f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

